import hashlib
import re
import textwrap
import uuid
from datetime import datetime, timedelta, timezone
from threading import Thread
from typing import Optional

import jwt
from cpl_core.database.context import DatabaseContextABC
from cpl_core.environment import ApplicationEnvironmentABC
from cpl_core.mailing import EMail, EMailClientABC
from cpl_core.utils import CredentialManager
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List
from cpl_translation import TranslatePipe
from flask import request

from bot_api.abc.auth_service_abc import AuthServiceABC
from bot_api.configuration.authentication_settings import AuthenticationSettings
from bot_api.configuration.frontend_settings import FrontendSettings
from bot_api.exception.service_error_code_enum import ServiceErrorCode
from bot_api.exception.service_exception import ServiceException
from bot_api.filter.auth_user_select_criteria import AuthUserSelectCriteria
from bot_api.logging.api_logger import ApiLogger
from bot_api.model.auth_user_dto import AuthUserDTO
from bot_api.model.auth_user_filtered_result_dto import AuthUserFilteredResultDTO
from bot_api.model.email_string_dto import EMailStringDTO
from bot_api.model.o_auth_dto import OAuthDTO
from bot_api.model.reset_password_dto import ResetPasswordDTO
from bot_api.model.token_dto import TokenDTO
from bot_api.model.update_auth_user_dto import UpdateAuthUserDTO
from bot_api.transformer.auth_user_transformer import AuthUserTransformer as AUT
from bot_data.abc.auth_user_repository_abc import AuthUserRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.auth_role_enum import AuthRoleEnum
from bot_data.model.auth_user import AuthUser
from bot_data.model.auth_user_users_relation import AuthUserUsersRelation

_email_regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'


class AuthService(AuthServiceABC):

    def __init__(
            self,
            env: ApplicationEnvironmentABC,
            logger: ApiLogger,
            bot: DiscordBotServiceABC,
            db: DatabaseContextABC,
            auth_users: AuthUserRepositoryABC,
            users: UserRepositoryABC,
            servers: ServerRepositoryABC,
            # mailer: MailThread,
            mailer: EMailClientABC,
            t: TranslatePipe,
            auth_settings: AuthenticationSettings,
            frontend_settings: FrontendSettings,

    ):
        AuthServiceABC.__init__(self)

        self._environment = env
        self._logger = logger
        self._bot = bot
        self._db = db
        self._auth_users = auth_users
        self._users = users
        self._servers = servers
        self._mailer = mailer
        self._t = t
        self._auth_settings = auth_settings
        self._frontend_settings = frontend_settings

    @staticmethod
    def _hash_sha256(password: str, salt: str) -> str:
        return hashlib.sha256(f'{password}{salt}'.encode('utf-8')).hexdigest()

    @staticmethod
    def _is_email_valid(email: str) -> bool:
        if re.fullmatch(_email_regex, email) is not None:
            return True

        return False

    def generate_token(self, user: AuthUser) -> str:
        token = jwt.encode(
            payload={
                'user_id': user.id,
                'email': user.email,
                'role': user.auth_role.value,
                'exp': datetime.now(tz=timezone.utc) + timedelta(days=self._auth_settings.token_expire_time),
                'iss': self._auth_settings.issuer,
                'aud': self._auth_settings.audience
            },
            key=CredentialManager.decrypt(self._auth_settings.secret_key)
        )

        return token

    def decode_token(self, token: str) -> dict:
        return jwt.decode(
            token,
            key=CredentialManager.decrypt(self._auth_settings.secret_key),
            issuer=self._auth_settings.issuer,
            audience=self._auth_settings.audience,
            algorithms=['HS256']
        )

    def get_decoded_token_from_request(self) -> dict:
        token = None
        if 'Authorization' in request.headers:
            bearer = request.headers.get('Authorization')
            token = bearer.split()[1]

        if token is None:
            raise ServiceException(ServiceErrorCode.Unauthorized, f'Token not set')

        return jwt.decode(
            token,
            key=CredentialManager.decrypt(self._auth_settings.secret_key),
            issuer=self._auth_settings.issuer,
            audience=self._auth_settings.audience,
            algorithms=['HS256']
        )

    def find_decoded_token_from_request(self) -> Optional[dict]:
        token = None
        if 'Authorization' in request.headers:
            bearer = request.headers.get('Authorization')
            token = bearer.split()[1]

        return jwt.decode(
            token,
            key=CredentialManager.decrypt(self._auth_settings.secret_key),
            issuer=self._auth_settings.issuer,
            audience=self._auth_settings.audience,
            algorithms=['HS256']
        ) if token is not None else None

    def _create_and_save_refresh_token(self, user: AuthUser) -> str:
        token = str(uuid.uuid4())
        user.refresh_token = token
        user.refresh_token_expire_time = datetime.now() + timedelta(days=self._auth_settings.refresh_token_expire_time)
        self._auth_users.update_auth_user(user)
        self._db.save_changes()
        return token

    def _send_link_mail(self, email: str, subject: str, message: str):
        url = self._frontend_settings.url
        if not url.endswith('/'):
            url = f'{url}/'

        self._mailer.connect()
        mail = EMail()
        mail.add_header('Mime-Version: 1.0')
        mail.add_header('Content-Type: text/plain; charset=utf-8')
        mail.add_header('Content-Transfer-Encoding: quoted-printable')
        mail.add_receiver(str(email))
        mail.subject = subject
        mail.body = textwrap.dedent(f"""{message}
        {self._t.transform('api.mail.automatic_mail').format(self._environment.application_name, self._environment.environment_name, self._environment.host_name)}
        """)

        thr = Thread(target=self._mailer.send_mail, args=[mail])
        thr.start()

    def _send_confirmation_id_to_user(self, user: AuthUser):
        url = self._frontend_settings.url
        if not url.endswith('/'):
            url = f'{url}/'

        self._send_link_mail(
            user.email,
            self._t.transform('api.auth.confirmation.subject').format(user.first_name, user.last_name),
            self._t.transform('api.auth.confirmation.message').format(url, user.confirmation_id)
        )

    def _send_forgot_password_id_to_user(self, user: AuthUser):
        url = self._frontend_settings.url
        if not url.endswith('/'):
            url = f'{url}/'

        self._send_link_mail(
            user.email,
            self._t.transform('api.auth.forgot_password.subject').format(user.first_name, user.last_name),
            self._t.transform('api.auth.forgot_password.message').format(url, user.forgot_password_id)
        )

    async def get_all_auth_users_async(self) -> List[AuthUserDTO]:
        result = self._auth_users.get_all_auth_users() \
            .select(lambda x: AUT.to_dto(x))
        return List(AuthUserDTO, result)

    async def get_filtered_auth_users_async(self, criteria: AuthUserSelectCriteria) -> AuthUserFilteredResultDTO:
        users = self._auth_users.get_filtered_auth_users(criteria)
        result = users.result.select(lambda x: AUT.to_dto(x))

        return AuthUserFilteredResultDTO(
            List(AuthUserDTO, result),
            users.total_count
        )

    async def get_auth_user_by_email_async(self, email: str, with_password: bool = False) -> AuthUserDTO:
        try:
            user = self._auth_users.get_auth_user_by_email(email)
            return AUT.to_dto(user, password=user.password if with_password else None)
        except Exception as e:
            self._logger.error(__name__, f'AuthUser not found', e)
            raise ServiceException(ServiceErrorCode.InvalidData, f'User not found {email}')

    async def find_auth_user_by_email_async(self, email: str) -> Optional[AuthUser]:
        user = self._auth_users.find_auth_user_by_email(email)
        return AUT.to_dto(user) if user is not None else None

    async def add_auth_user_async(self, user_dto: AuthUserDTO):
        db_user = self._auth_users.find_auth_user_by_email(user_dto.email)
        if db_user is not None:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'User already exists')

        user = AUT.to_db(user_dto)
        if self._auth_users.get_all_auth_users().count() == 0:
            user.auth_role = AuthRoleEnum.admin

        user.password_salt = uuid.uuid4()
        user.password = self._hash_sha256(user_dto.password, user.password_salt)
        if not self._is_email_valid(user.email):
            raise ServiceException(ServiceErrorCode.InvalidData, 'Invalid E-Mail address')

        try:
            user.confirmation_id = uuid.uuid4()
            self._auth_users.add_auth_user(user)
            self._send_confirmation_id_to_user(user)
            self._db.save_changes()
            self._logger.info(__name__, f'Added auth user with E-Mail: {user_dto.email}')
        except Exception as e:
            self._logger.error(__name__, f'Cannot add user with E-Mail {user_dto.email}', e)
            raise ServiceException(ServiceErrorCode.UnableToAdd, "Invalid E-Mail")

    async def add_auth_user_by_oauth_async(self, dto: OAuthDTO):
        db_user = self._auth_users.find_auth_user_by_email(dto.user.email)

        if db_user is None:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'User not found')

        if db_user.oauth_id != dto.oauth_id:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'Wrong OAuthId')

        try:
            db_user.first_name = dto.user.first_name
            db_user.last_name = dto.user.last_name
            db_user.password_salt = uuid.uuid4()
            db_user.password = self._hash_sha256(dto.user.password, db_user.password_salt)
            db_user.oauth_id = None
            db_user.confirmation_id = uuid.uuid4()
            self._send_confirmation_id_to_user(db_user)
            self._auth_users.update_auth_user(db_user)
            self._logger.info(__name__, f'Added auth user with E-Mail: {dto.user.email}')
        except Exception as e:
            self._logger.error(__name__, f'Cannot add user with E-Mail {dto.user.email}', e)
            raise ServiceException(ServiceErrorCode.UnableToAdd, "Invalid E-Mail")

        self._db.save_changes()

    async def add_auth_user_by_discord_async(self, user_dto: AuthUserDTO, dc_id: int) -> OAuthDTO:
        db_auth_user = self._auth_users.find_auth_user_by_email(user_dto.email)

        # user exists
        if db_auth_user is not None and db_auth_user.users.count() > 0:
            # raise ServiceException(ServiceErrorCode.InvalidUser, 'User already exists')
            self._logger.debug(__name__, f'Discord user already exists')
            return OAuthDTO(AUT.to_dto(db_auth_user), None)

        # user exists but discord user id not set
        elif db_auth_user is not None and db_auth_user.users.count() == 0:
            self._logger.debug(__name__, f'Auth user exists but not linked with discord')
            # users = self._users.get_users_by_discord_id(user_dto.user_id)
            # add auth_user to user refs
            db_auth_user.oauth_id = None

        else:
            # user does not exists
            self._logger.debug(__name__, f'Auth user does not exist')
            try:
                user_dto.user_id = self._users.get_users_by_discord_id(user_dto.user_id).single().user_id
            except Exception as e:
                self._logger.error(__name__, f'User not found')
                user_dto.user_id = None

            await self.add_auth_user_async(user_dto)
            db_auth_user = self._auth_users.get_auth_user_by_email(user_dto.email)
            db_auth_user.oauth_id = uuid.uuid4()

        for g in self._bot.guilds:
            member = g.get_member(int(dc_id))
            if member is None:
                continue

            server = self._servers.get_server_by_discord_id(g.id)
            users = self._users.get_users_by_discord_id(dc_id)
            for user in users:
                if user.server.server_id != server.server_id:
                    continue
                self._auth_users.add_auth_user_user_rel(AuthUserUsersRelation(db_auth_user, user))

        self._auth_users.update_auth_user(db_auth_user)
        self._db.save_changes()
        return OAuthDTO(AUT.to_dto(db_auth_user), db_auth_user.oauth_id)

    async def update_user_async(self, update_user_dto: UpdateAuthUserDTO):
        if update_user_dto is None:
            raise ServiceException(ServiceErrorCode.InvalidData, f'User is empty')

        if update_user_dto.auth_user is None:
            raise ServiceException(ServiceErrorCode.InvalidData, f'Existing user is empty')

        if update_user_dto.new_auth_user is None:
            raise ServiceException(ServiceErrorCode.InvalidData, f'New user is empty')

        if not self._is_email_valid(update_user_dto.auth_user.email) or not self._is_email_valid(update_user_dto.new_auth_user.email):
            raise ServiceException(ServiceErrorCode.InvalidData, f'Invalid E-Mail')

        user = self._auth_users.find_auth_user_by_email(update_user_dto.auth_user.email)
        if user is None:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'User not found')

        if user.confirmation_id is not None:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'E-Mail not confirmed')

        # update first name
        if update_user_dto.new_auth_user.first_name is not None and update_user_dto.auth_user.first_name != update_user_dto.new_auth_user.first_name:
            user.first_name = update_user_dto.new_auth_user.first_name

        # update last name
        if update_user_dto.new_auth_user.last_name is not None and update_user_dto.new_auth_user.last_name != '' and \
                update_user_dto.auth_user.last_name != update_user_dto.new_auth_user.last_name:
            user.last_name = update_user_dto.new_auth_user.last_name

        # update E-Mail
        if update_user_dto.new_auth_user.email is not None and update_user_dto.new_auth_user.email != '' and update_user_dto.auth_user.email != update_user_dto.new_auth_user.email:
            user_by_new_e_mail = self._auth_users.find_auth_user_by_email(update_user_dto.new_auth_user.email)
            if user_by_new_e_mail is not None:
                raise ServiceException(ServiceErrorCode.InvalidUser, 'User already exists')
            user.email = update_user_dto.new_auth_user.email

        update_user_dto.auth_user.password = self._hash_sha256(update_user_dto.auth_user.password, user.password_salt)
        if update_user_dto.auth_user.password != user.password:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'Wrong password')

        # update password
        if update_user_dto.new_auth_user.password is not None and self._hash_sha256(update_user_dto.new_auth_user.password, user.password_salt) != user.password:
            user.password_salt = uuid.uuid4()
            user.password = self._hash_sha256(update_user_dto.new_auth_user.password, user.password_salt)

        self._auth_users.update_auth_user(user)
        self._db.save_changes()

    async def update_user_as_admin_async(self, update_user_dto: UpdateAuthUserDTO):
        if update_user_dto is None:
            raise ServiceException(ServiceErrorCode.InvalidData, f'User is empty')

        if update_user_dto.auth_user is None:
            raise ServiceException(ServiceErrorCode.InvalidData, f'Existing user is empty')

        if update_user_dto.new_auth_user is None:
            raise ServiceException(ServiceErrorCode.InvalidData, f'New user is empty')

        if not self._is_email_valid(update_user_dto.auth_user.email) or not self._is_email_valid(update_user_dto.new_auth_user.email):
            raise ServiceException(ServiceErrorCode.InvalidData, f'Invalid E-Mail')

        user = self._auth_users.find_auth_user_by_email(update_user_dto.auth_user.email)
        if user is None:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'User not found')

        if user.confirmation_id is not None and update_user_dto.new_auth_user.is_confirmed:
            user.confirmation_id = None
        elif user.confirmation_id is None and not update_user_dto.new_auth_user.is_confirmed:
            user.confirmation_id = uuid.uuid4()
        # else
        #     raise ServiceException(ServiceErrorCode.InvalidUser, 'E-Mail not confirmed')

        # update first name
        if update_user_dto.new_auth_user.first_name is not None and update_user_dto.auth_user.first_name != update_user_dto.new_auth_user.first_name:
            user.first_name = update_user_dto.new_auth_user.first_name

        # update last name
        if update_user_dto.new_auth_user.last_name is not None and update_user_dto.new_auth_user.last_name != '' and update_user_dto.auth_user.last_name != update_user_dto.new_auth_user.last_name:
            user.last_name = update_user_dto.new_auth_user.last_name

        # update E-Mail
        if update_user_dto.new_auth_user.email is not None and update_user_dto.new_auth_user.email != '' and update_user_dto.auth_user.email != update_user_dto.new_auth_user.email:
            user_by_new_e_mail = self._auth_users.find_auth_user_by_email(update_user_dto.new_auth_user.email)
            if user_by_new_e_mail is not None:
                raise ServiceException(ServiceErrorCode.InvalidUser, 'User already exists')
            user.email = update_user_dto.new_auth_user.email

        # update password
        if update_user_dto.change_password and user.password != self._hash_sha256(update_user_dto.new_auth_user.password, user.password_salt):
            user.password_salt = uuid.uuid4()
            user.password = self._hash_sha256(update_user_dto.new_auth_user.password, user.password_salt)

        # update role
        if user.auth_role == update_user_dto.auth_user.auth_role and user.auth_role != update_user_dto.new_auth_user.auth_role:
            user.auth_role = update_user_dto.new_auth_user.auth_role

        self._auth_users.update_auth_user(user)
        self._db.save_changes()

    async def delete_auth_user_by_email_async(self, email: str):
        try:
            user = self._auth_users.get_auth_user_by_email(email)
            self._auth_users.delete_auth_user(user)
            self._db.save_changes()
        except Exception as e:
            self._logger.error(__name__, f'Cannot delete user', e)
            raise ServiceException(ServiceErrorCode.UnableToDelete, f'Cannot delete user by mail {email}')

    async def delete_auth_user_async(self, user_dto: AuthUser):
        try:
            self._auth_users.delete_auth_user(AUT.to_db(user_dto))
            self._db.save_changes()
        except Exception as e:
            self._logger.error(__name__, f'Cannot delete user', e)
            raise ServiceException(ServiceErrorCode.UnableToDelete, f'Cannot delete user by mail {user_dto.email}')

    def verify_login(self, token_str: str) -> bool:
        try:
            token = self.decode_token(token_str)
            if token is None or 'email' not in token:
                raise ServiceException(ServiceErrorCode.InvalidData, 'Token invalid')

            user = self._auth_users.find_auth_user_by_email(token['email'])
            if user is None:
                raise ServiceException(ServiceErrorCode.InvalidData, 'Token expired')
        except Exception as e:
            self._logger.error(__name__, f'Token invalid', e)
            return False

        return True

    async def login_async(self, user_dto: AuthUser) -> TokenDTO:
        if user_dto is None:
            raise ServiceException(ServiceErrorCode.InvalidData, 'User not set')

        db_user = self._auth_users.find_auth_user_by_email(user_dto.email)
        if db_user is None:
            raise ServiceException(ServiceErrorCode.InvalidUser, f'User not found')

        user_dto.password = self._hash_sha256(user_dto.password, db_user.password_salt)
        if db_user.password != user_dto.password:
            raise ServiceException(ServiceErrorCode.InvalidUser, 'Wrong password')

        token = self.generate_token(db_user)
        refresh_token = self._create_and_save_refresh_token(db_user)
        if db_user.forgot_password_id is not None:
            db_user.forgot_password_id = None

        self._db.save_changes()
        return TokenDTO(token, refresh_token)

    async def refresh_async(self, token_dto: TokenDTO) -> TokenDTO:
        if token_dto is None:
            raise ServiceException(ServiceErrorCode.InvalidData, f'Token not set')

        try:
            token = self.decode_token(token_dto.token)
            if token is None or 'email' not in token:
                raise ServiceException(ServiceErrorCode.InvalidData, 'Token invalid')

            user = self._auth_users.get_auth_user_by_email(token['email'])
            if user is None or user.refresh_token != token_dto.refresh_token or user.refresh_token_expire_time <= datetime.now():
                raise ServiceException(ServiceErrorCode.InvalidData, 'Token expired')

            return TokenDTO(self.generate_token(user), self._create_and_save_refresh_token(user))
        except Exception as e:
            self._logger.error(__name__, f'Refreshing token failed', e)
            return TokenDTO('', '')

    async def revoke_async(self, token_dto: TokenDTO):
        if token_dto is None or token_dto.token is None or token_dto.refresh_token is None:
            raise ServiceException(ServiceErrorCode.InvalidData, 'Token not set')

        try:
            token = self.decode_token(token_dto.token)

            user = self._auth_users.get_auth_user_by_email(token['email'])
            if user is None or user.refresh_token != token_dto.refresh_token or user.refresh_token_expire_time <= datetime.now():
                raise ServiceException(ServiceErrorCode.InvalidData, 'Token expired')

            user.refresh_token = None
            self._auth_users.update_auth_user(user)
            self._db.save_changes()
        except Exception as e:
            self._logger.error(__name__, f'Refreshing token failed', e)

    async def confirm_email_async(self, id: str) -> bool:
        user = self._auth_users.find_auth_user_by_confirmation_id(id)
        if user is None:
            return False

        user.confirmation_id = None
        self._auth_users.update_auth_user(user)
        self._db.save_changes()
        return True

    async def forgot_password_async(self, email: str):
        user = self._auth_users.find_auth_user_by_email(email)
        if user is None:
            return

        user.forgot_password_id = uuid.uuid4()
        self._auth_users.update_auth_user(user)
        self._send_forgot_password_id_to_user(user)
        self._db.save_changes()

    async def confirm_forgot_password_async(self, id: str) -> EMailStringDTO:
        user = self._auth_users.find_auth_user_by_forgot_password_id(id)
        return EMailStringDTO(user.email)

    async def reset_password_async(self, rp_dto: ResetPasswordDTO):
        user = self._auth_users.find_auth_user_by_forgot_password_id(rp_dto.id)
        if user is None:
            raise ServiceException(ServiceErrorCode.InvalidUser, f'User by forgot password id {rp_dto.id} not found')

        if user.confirmation_id is not None:
            raise ServiceException(ServiceErrorCode.InvalidUser, f'E-Mail not confirmed')

        if user.password is None or rp_dto.password == '':
            raise ServiceException(ServiceErrorCode.InvalidData, f'Password not set')

        user.password_salt = uuid.uuid4()
        user.password = self._hash_sha256(rp_dto.password, user.password_salt)
        user.forgot_password_id = None
        self._auth_users.update_auth_user(user)
        self._db.save_changes()
